#include <port.h>

/* INCLUDES */
#include <math.h>
#include <string.h>
#include "uc.h"
#include "imgformat.h"
#include "memory.h"


/* Global variables definition */
extern PGMImage        *pgmimage ;
extern char             outfile[MAXPATHLEN] ;

extern float            sigma;
extern int              mask;
extern float            low_threshold;
extern float            high_threshold;
extern int              XSIZE;
extern int              YSIZE;

extern int              Sub_pixel;
extern int              NUM_EDGELS;
extern int              NUM_EDGES;
extern int              NUM_JCTS;

extern unsigned short   *Record_image ;
extern unsigned short   *image;
extern unsigned short   *timage;

extern edgel            **edgels ;
extern edgel            *tedgels ;

extern edge             *edges ;
extern edge             *tedges ;


/* A Macro that aids referencing our various pointers */

#define DEREF(base, xsize, x, y) (*((base)+(((int) (y+0.5))*(xsize))+((int) (x+0.5))))


/* Global offsets used to direct one around the compass from the 
   current location    */
double x_offset[MAXDIRECT] = { -1.0, 0.0, 1.0,
                               -1.0,     -1.0,
                                0.0, 1.0, 1.0 } ;

double y_offset[MAXDIRECT] = { 1.0,  1.0,  1.0,
                               0.0,       -1.0,
                              -1.0, -1.0,  0.0 } ;


void gaussian_vectors() ;
void add_edgel() ;
void set_edgel_high() ;
void local_extrema() ;
void join_edges() ;
void add_edges() ;
void walk() ;
double newx() ;
double newy() ;
int num_neighbors() ;
int junction_done() ;
void delete_junction() ;
void set_junction() ;
junction_list *get_junction() ;
edgel *get_edgel() ;
edgel *get_edgel_ptr() ;
void write_file() ;
void localize() ;
edgel **new_edgel_list() ;
int edgel_key() ;


/*************************************************************************/
/*                                                                       */
/* CONVOLVE.C                                                            */
/*                                                                       */
/* DESCRIPTION:                                                          */
/*   This routine will perform a Gaussian convolution given the          */
/*   dimensions of the image and the mask.  Besides returning the X and  */
/*   Y gradients, it will also return the magnitude of the gradient to   */
/*   the user.                                                           */
/*                                                                       */
/* COMMENTS:                                                             */
/*   Use the gloabal variable "image" to access the row-major stored,    */
/*   unsigned-short pixel info.                                          */ 
/*                                                                       */
/*************************************************************************/
void convolve(unsigned char **src_image,
              int mask_size,	/* dimension of gaussian mask */
              double sigma,	/* gaussian standard deviation */
              double **gx,	/* X gradient of convolved image */
              double **gy,	/* Y gradient of convolved image */
              double **mag)	/* magnitude of gradient */
  {
  double		*gp_array ;	/* ptr to the 2-D gaussian prime array */
  double		*g_vec ;	/* ptr to the gaussian vector */
  int			half_mask_size ;/* half the mask size (rounded up) */
  register int		idx ;		/* looping index */
  unsigned char		*image_ptr ;	/* temp pointer to image data */
  int			image_size ;	/* the size of the image in bytes */
  double		*int_gx ;	/* intermediate x gradient array */
  double		*int_gy ;	/* intermediate y gradient array */
  double		int_gx_val ;	/* temp value var from int_gx */
  double		int_gy_val ;	/* temp value var from int_gy */
  int			pixel ;		/* an intensity value of a point in the image */
  int			pixel_offset ;	/* index into gaussian prime array */
  register double	product ;	/* product of gaussian and an image point */
  register double	product_gx ;	/* temp product var */
  register double	product_gy ;	/* temp product var */
  int			start_x ;	/* left   edge for convolution */
  int			start_y ;	/* top    edge for convolution */
  int			end_x ;	        /* right  edge for convolution */
  int			end_y ;	        /* bottom edge for convolution */
  register double	*temp_gx ;	/* temporary gradients ptr */
  register double	*temp_gy ;
  register double	*temp_int_gx ;
  register double	*temp_int_gy ;
  register double	*temp_mag ;	/* temporary magnitude ptr */
  register int		x_pos ;		/* looping index */
  register int		y_pos ;		/* looping index */

  /* Make sure that the mask size is odd so that there is a definite
     center pixel, with minimum width of 3 */
  if ((mask_size % 2) == 0)
    mask_size++ ;
  mask_size = MAX(mask_size, 3) ;
   
  /* Build gaussian vector and the first-derivitive of gaussian
     filter array.  The latter is a 2-D array with rows equal
     to the gaussian prime vector multiplied by the index of the row.
     This means that all possible products are pre-calculated for
     convolution with the pixel values. */
  gaussian_vectors(mask_size, sigma, &gp_array, &g_vec) ;

  /* Initialize the time-saving variables */
  half_mask_size = (mask_size + 1) / 2 ;
  image_size = XSIZE * YSIZE ;

  /* Allocate the gradient arrays: both the official x and y
     gradient arrays, as well as the intermediate gradient array space */
  (*gx) = dvector(image_size) ;
  (*gy) = dvector(image_size) ;
  int_gx = dvector(image_size) ;
  int_gy = dvector(image_size) ;

  /* Initialize gradient arrays to zero */
  temp_gx = *gx ;
  temp_gy = *gy ;
  temp_int_gx = int_gx ;
  temp_int_gy = int_gy ;

  for (idx = 0 ; idx < image_size ; idx++)
    {
    temp_gx[idx] = 0.0 ;
    temp_gy[idx] = 0.0 ;
    temp_int_gx[idx] = 0.0 ;
    temp_int_gy[idx] = 0.0 ;
    } ;

  /* Traverse the image pixel by pixel and perform the convolution
     with the gaussian prime mask, generating the first stage of the
     the gx and gy arrays.  Then convolve in the orthogonal direction
     with a normal gaussian for smoothing purposes. */
  start_x = 0 ;
  start_y = 0 ;
  end_x   = XSIZE ;
  end_y   = YSIZE ;
/*  image_ptr   = image  + (start_y * XSIZE) + start_x ; */
  temp_int_gx = int_gx + (start_y * XSIZE) + start_x ;
  temp_int_gy = int_gy + (start_y * XSIZE) + start_x ;
   
  for (y_pos = start_y ; y_pos < end_y ; y_pos++)
    {
    for (x_pos = start_x ; x_pos < end_x ; x_pos++)
      {
      /* Calculate the index into the gaussian prime array based on the 
         intensity of the current pixel.  This gives a pointer to the 
         start of the appropriate pre-multiplied gaussian prime vector. */
      pixel_offset = (int) src_image[y_pos][x_pos] * half_mask_size ;
/*      pixel_offset = ((int)(*image_ptr)) * half_mask_size ; */
      product = gp_array[pixel_offset] ;
      
      /* Handle the first element of the mask (the middle element of
         the gaussian derivative) */
      *temp_int_gx += product ;
      *temp_int_gy += product ;
      
      /* Add in the rest of the pre-multiplied values */
      for(idx = 1 ; idx < half_mask_size ; idx++)
        {
        product = gp_array[pixel_offset + idx] ;
        /* Contribute to a pixel to the right in intermediate gx */
        if (x_pos + idx < XSIZE)
          *(temp_int_gx + idx) += product ;	/* A pixel that far to the right exists, so add it in */
        
        /* Contribute to a pixel to the left in intermediate gx */
        if (x_pos - idx >= 0)
          *(temp_int_gx - idx) -= product ;	/* A pixel that far to the left exists, so subtract it in */
        
        /* Contribute to a pixel below in intermediate gy */
        if (y_pos + idx < YSIZE)
          *(temp_int_gy + (idx * XSIZE)) += product ;	/* A pixel that far below exists, so add it in */

        /* Contribute to a pixel above in intermediate gy */
        if (y_pos - idx >= 0)
          *(temp_int_gy - (idx * XSIZE)) -= product ;	/* A pixel that far above exists, so subtract it in */
        }
      /* Go to next pixel in both intermediate gradient arrays and in the image itself */
      temp_int_gx++ ;
      temp_int_gy++ ;
/*      image_ptr++ ; */
      }
    }

   
  /* Now perform complementary convolution with a normal gaussian. That is,
     the x gradient array gets convolved with a vertical gaussian, and the 
     y gradient array gets convolved with a horizontal gaussian. */
  start_x = 0 ;
  start_y = 0 ;
  end_x   = XSIZE ;
  end_y   = YSIZE ;
  temp_int_gx = int_gx ;
  temp_int_gy = int_gy ;
  temp_gx     = *gx ;
  temp_gy     = *gy ;
   
  for (y_pos = start_y ; y_pos < end_y ; y_pos++)
    {
    for (x_pos = start_x ; x_pos < end_x ; x_pos++)
      {
      /* Retrieve the calculated intermediate values for use later */
      int_gx_val = *temp_int_gx ;
      int_gy_val = *temp_int_gy ;
      
      /* Handle the first element of the mask (the middle element of the gaussian) */
      product_gx = g_vec[0] * int_gx_val ;
      product_gy = g_vec[0] * int_gy_val ;
      
      *temp_gx += product_gx ;
      *temp_gy += product_gy ;
      
      /* Multiply in the rest of the gaussian values */
      for(idx = 1 ; idx < half_mask_size ; idx++)
        {
        product_gx = g_vec[idx] * int_gx_val ;
        product_gy = g_vec[idx] * int_gy_val ;
        
        /* Contribute to a pixel to the right in gy */
        if (x_pos + idx < XSIZE)
          *(temp_gy + idx) += product_gy ;	/* A pixel that far to the right exists, so add it in */
        
        /* Contribute to a pixel to the left in gy */
        if (x_pos - idx >= 0)
          *(temp_gy - idx) += product_gy ;	/* A pixel that far to the left exists, so add it in */
        
        /* Contribute to a pixel below in gx */
        if (y_pos + idx < YSIZE)
          *(temp_gx + (idx * XSIZE)) += product_gx ;	/* A pixel that far below exists, so add it in */
        
        /* Contribute to a pixel above in gx */
        if (y_pos - idx >= 0)
          *(temp_gx - (idx * XSIZE)) += product_gx ;	/* A pixel that far above exists, so add it in */
        }
      
      /* Go to the next pixel in both the intermediate and the final gradient arrays */
      temp_int_gx++ ;
      temp_int_gy++ ;
      temp_gx++ ;
      temp_gy++ ;
      }
    }
  
  /* Free up the intermediate gradient arrays, and allocate the magnitude array */
  free_dvector(int_gx, image_size) ;
  free_dvector(int_gy, image_size) ;
   
  *mag = dvector(image_size) ;
  
  /* Reset the temporary pointers to the beginning of the gradient
     and magnitude arrays so that the magnitude can be computed using
     the distance. */
  temp_gx  = *gx ;
  temp_gy  = *gy ;
  temp_mag = *mag ;

  for (idx = image_size, temp_gx = *gx, temp_gy = *gy, temp_mag = *mag ; 
       idx > 0 ; 
       idx--, temp_gx++, temp_gy++, temp_mag++)
    *temp_mag = hypot(*temp_gx, *temp_gy ) ;	/* Calculate the magnitude */
   
  /* Clean up and return to the user */
  free(gp_array) ;
  free(g_vec) ;
  }


/*************************************************************************/
/*                                                                       */
/* GAUSSIAN_VECTORS                                                      */
/*                                                                       */
/* DESCRIPTION:                                                          */
/*   Build the array of numbers used to perform the convolution.         */
/*   This is a two dimensional array with 256 rows (0 to 255).  In       */
/*   each row is the right half of the convolution mask multiplied       */
/*   by the row number.                                                  */
/*                                                                       */
/* NOTE:                                                                 */
/*   gaussian   = g(x)                                                   */
/*              =  1/(sigma   * sqrt(2*PI)) * exp( -x^2 / 2*sigma^2 )    */
/*   derivative = g'(x)                                                  */
/*              = -x/(sigma^3 * sqrt(2*PI)) * exp( -x^2 / 2*sigma^2 )    */
/*                                                                       */
/*************************************************************************/
void gaussian_vectors(int mask_size,
                      double sigma,
                      double **gp_array_ptr,
                      double ** g_vec_ptr)
  {
  double	*gp_vec ;		/* pointer to the first-derivative of the gaussian vector (gaussian prime) */
  int		half_mask_size ;	/* half of the mask size (rounded up) */
  double	inv_sigma_rt_2_pi ;	/* reciprocal of (sigma times the square root of 2 * PI) */
  double	sigma_sq ;		/* square of sigma */
  double	*temp_ptr ;		/* temp ptr to the 2-D array */
  double	two_sigma_sq ;		/* two times square of sigma */
  double	x_val ;			/* loop index for calculating the gaussians */
  double	*g_vec, *gp_array ;
  int		idx ;			/* looping index */
  int		col_idx ;		/* looping index */
  int		row_idx ;		/* looping index */
   
  /* Make sure the mask size is odd */
  if (( mask_size % 2 ) == 0 )
    fprintf( stderr, "\nERROR: Mask size of %d is even!\n\n", mask_size ) ;

  half_mask_size = (mask_size + 1) / 2 ;
   
  /* Allocate storage for the gaussians */
  gp_vec = dvector(half_mask_size) ;
  g_vec = dvector(half_mask_size) ;
  gp_array = dvector(half_mask_size * 256) ;

  /* Initialize variables we use often */
  sigma_sq          = sigma * sigma ;
  two_sigma_sq      = 2.0 * sigma_sq ;
  inv_sigma_rt_2_pi = 1.0 / (sigma * sqrt(2.0 * PI)) ;
   
  /* Fill in the gaussian prime vector and the gaussian vector */
  /* Start at an x value of 0 because mask size is odd*/
  for (idx = 0, x_val = 0.0 ; idx < half_mask_size ; idx++, x_val++)
    {
    g_vec[idx] = inv_sigma_rt_2_pi * exp(-(x_val * x_val) / two_sigma_sq) ;
    gp_vec[idx] = (-x_val / sigma_sq) * g_vec[idx] ;
    } 
   
  /* Calculate the 2-D array from the gaussian-prime vector */
  for (row_idx = 255 ; row_idx >= 0 ; row_idx--)
    {
    /* Get a pointer to the start of the row */
    temp_ptr = gp_array + row_idx * half_mask_size ;
    for (col_idx = half_mask_size - 1 ; col_idx >= 0 ; col_idx--)
      temp_ptr[col_idx] = (double) row_idx * gp_vec[col_idx] ;
    }

  /* Free the gaussian derivative vector */
  free_dvector(gp_vec, half_mask_size) ;
  
  *g_vec_ptr = g_vec ;
  *gp_array_ptr = gp_array ;
  } 


/*************************************************************************/
/*                                                                       */
/* ADD_EDGEL                                                             */
/*                                                                       */
/* PURPOSE:                                                              */
/*   This will create a new edgel and add it to the hash-table of        */
/*   edgels.                                                             */
/*                                                                       */
/*************************************************************************/
void add_edgel(edgel *edgel_table[],	/* the hash table of edgels */
               double x_pos,		/* Input edgel coordinates */
               double y_pos,
               double p_grad,		/* Input gradient values */
               double q_grad,
               int high,
               double xreal,		/* Localized X,Y points in */
               double yreal)		/* Cartesian space. */
  {
  edgel		**bin_ptr ;	/* ptr to the start of a list edgels
  				   in a bin of the hash-table */
  edgel		*new_edgel ;	/* ptr to the new edgel */
  
  /* Allocate space for the new edgel */
  new_edgel = (edgel *) calloc(1, sizeof(edgel)) ;
  
  /* Fill in the slots in the record */
  new_edgel->pos.x  = x_pos ;
  new_edgel->pos.y  = y_pos ;
  new_edgel->grad.p = p_grad ;
  new_edgel->grad.q = q_grad ;
  new_edgel->loc.x  = xreal ;
  new_edgel->loc.y  = yreal ;
  new_edgel->high   = high ;
  
  /* Find which bin in the hash-table to use */
  bin_ptr = &(edgel_table[edgel_key(x_pos, y_pos)]) ;
  
  /* Set the next pointer to point to the front of the bin's edgel list */
  new_edgel->next = *bin_ptr ;
  
  /* The first edgel in the bin is now new_edgel */
  *bin_ptr = new_edgel ;
  }


/*************************************************************************/
/*                                                                       */
/* SET_EDGEL_HIGH                                                        */
/*                                                                       */
/* PURPOSE:                                                              */
/*   This finds the edgel with position (x_pos, y_pos) and sets its      */
/*   "high" field.  This "promotes" a low edgel that is connected to a   */
/*   junction that was not already marked as being visited.              */
/*                                                                       */
/*************************************************************************/
void set_edgel_high(edgel *edgel_table[],	/* ptr to the edgel table */
                    double x_pos,		/* Input edgel coordinates */
                    double y_pos)
  {
  edgel		*edgel_ptr ;		/* temp ptr to the requested edgel */
  
  /* Get the pointer to the correct edgel */
  edgel_ptr = get_edgel_ptr(edgel_table, x_pos, y_pos) ;
  
  /* Set the "high" flag */
  if (edgel_ptr == NULL)
    fprintf(stderr, "\n\nERROR:  Could not set (%.1f, %.1f) high!\n",
            x_pos, y_pos ) ;
  else
    edgel_ptr->high = TRUE ;
  }


/*************************************************************************/
/*                                                                       */
/* LOCAL_EXTREMA                                                         */
/*                                                                       */
/* PURPOSE:                                                              */
/*   This routine looks through the image for gradient values that are   */
/*   above the low threshold value.  The values that are above the high  */
/*   threshold are placed in a linked list so that once we start to join */
/*   the edges together, we can 'jump' onto a valid starting pixel and   */
/*   then look for the next edge point.  The idea behind this is that we */
/*   want to avoid traversing the entire image again.  This will return  */
/*   both a linked list of starting edge points above the high threshold */
/*   as well as a new record-image with non edge points turned 'off'.    */
/*                                                                       */
/*************************************************************************/
void local_extrema(int mask,		/* Mask */
                   double *gx,		/* X Gradient of convolved image */
                   double *gy,		/* Y Gradient of convolved image */
                   double *mag,		/* Magnitude of gradient */
                   double low_t,	/* low and high threshold values */
                   double high_t)
  {
  double		gu, gd ;	/* The gradient in the up    */
					/* and down direction.       */
  double		direction ;	/* tangent of the gradient */
  int			idx ;		/* for-loop index */
  int			halfM ;		/* Cut the mask into half    */
  int			x, y ;		/* Cartiesain coords of the image */
  unsigned short	*temp_ptr ;	/* temp ptr to the record image */
  register double	*tmag ;		/* Temp pointer into the magnitude*/
  register double	*tgx ;
  register double	*tgy ;
  double		xreal ;		/* Sub-pixel localization of both */
  double		yreal ;		/* the X&Y Cartiesain coordinates */
  


  /* Initialize the edgels structure (this is a global variable) */
  edgels = new_edgel_list() ;

  /* Divide the mask into half */
  halfM = (mask + 1) / 2 ;

  /* Initialize coordinates of array */
  x = 0 ;
  y = 0 ;

  /* Set the temporary pointer to the beginning of the gradient arrays */
  tgx = gx;
  tgy = gy;

  /* Loop through the magnitude matrix, looking for elements which exceed 
     the lower threshold.
     WARNING!!!: no boundry checking is done by DEREF for (x, y) locations
     that may be off the image.  This error is ignored for now becasue with
     FAST_CONV defined we are guaranteed to have 0 magnitude gradients
     within one mask width of the periphery. */
  for (idx = XSIZE * YSIZE ; idx > 0 ; idx--, tgx++, tgy++)
    {
    if (DEREF( mag, XSIZE, x, y ) >= low_t && 
        x < pgmimage->ImageWidth - 1       &&
        y < pgmimage->ImageHeight - 1      &&
        x > 0                              &&
        y > 0)
      {
      /* Find out which way the gradient points.  That is, calculate
         the tangent of the gradient at that coordinate. */
      direction = ( *tgx != 0.0 ) ? ( *tgy / *tgx ) : 0.0 ;
      
      if (*tgx == 0.0)
        {
        /* Angle == 90 degrees */
        gu = DEREF(mag, XSIZE, x, y+1) ;
        gd = DEREF(mag, XSIZE, x, y-1) ;
        }
      else
        {
        if (direction < 0)
          {
          /* Upper Left or Lower Right Side (i.e. quadrants II and IV) */
          /* Angle is < 0 degrees */
          if (direction > -1.0)
            {
            /*  -45 < Angle < 0  */
            gd = (1.0 + direction) * DEREF(mag, XSIZE, x-1, y) -
                  direction * DEREF(mag, XSIZE, x-1, y+1) ;
            gu = (1.0 + direction) * DEREF(mag, XSIZE, x+1, y) -
                  direction * DEREF(mag, XSIZE, x+1, y-1) ;
            }
          else
            {
            /*  -90 < Angle <= -45  */
            gd = (1.0+1.0/direction) * DEREF(mag, XSIZE, x, y+1) -
                 (1.0/direction) * DEREF(mag, XSIZE, x-1, y+1) ;
            gu = (1.0+1.0/direction) * DEREF(mag, XSIZE, x, y-1) - 
                 (1.0/direction) * DEREF(mag, XSIZE, x+1, y-1) ;
            }
          }
        else
          {
          /* Right Side (i.e. quadrants I and III) */
          /* Angle is >= 0 degrees */
          if (direction < 1.0)
            {
            /*  0 <= Angle < 45  */
            gu = (1.0-direction) * DEREF(mag, XSIZE, x+1, y) + 
                  direction * DEREF(mag, XSIZE, x+1, y+1) ;
            gd = (1.0-direction) * DEREF(mag, XSIZE, x-1, y) + 
                  direction * DEREF(mag, XSIZE, x-1, y-1) ;
            }
          else
            {
            /*  45 <= Angle < 90  */
            gu = (1.0-1.0/direction) * DEREF(mag, XSIZE, x, y+1) + 
                 (1.0/direction) * DEREF(mag, XSIZE, x+1, y+1) ;
            gd = (1.0-1.0/direction) * DEREF(mag, XSIZE, x, y-1) + 
                 (1.0/direction) * DEREF(mag, XSIZE, x-1, y-1) ;
            }
          }
        } /* end of case for angles != 90 degrees */
      
      /* If this is a local maximum then add it to the list of edgels */
      if ( (DEREF(mag, XSIZE, x, y) > gu)  &&
           (DEREF(mag, XSIZE, x, y) >= gd) &&
           (x > halfM)                     &&
           (y > halfM)                     &&
           (x < XSIZE - halfM)             &&
           (y < YSIZE - halfM) )
        {
        /* Calculate sub-pixel X & Y coordinates of the edgel */
        localize(DEREF(mag, XSIZE, x, y), gu, gd, direction, 
                 (double) x,(double) y, 
                 DEREF(gx, XSIZE, x, y), 
                 DEREF(gy, XSIZE, x, y), &xreal, &yreal) ;
        
        /* Create and add an edgel to the edgel table */
        add_edgel(edgels, (double) x, (double) y,
                  DEREF(gx, XSIZE, x, y), DEREF(gy, XSIZE, x, y),
                  (DEREF(mag, XSIZE, x, y) >= high_t), xreal, yreal) ;
        
        /* Increment the edgel count */
        NUM_EDGELS++ ;
        } /* Done with adding an edgel to the list */
      } /* End of checking for a potential edgel */
    
    /* Increment the x and y coordinates of the present position in 
       the matrix. */
    x++ ;
    if (x >= XSIZE)
      {
      x = 0 ;
      y++ ;
      }
    } /* End of looping through the image */
    
  /* Make an array for keeping track of the edgels */
  Record_image = (unsigned short *) calloc(XSIZE * YSIZE,
                                           sizeof(unsigned short)) ;
  
  /* Zero out everything below the low threshold */
  temp_ptr = Record_image ;
  for (idx = XSIZE * YSIZE ; idx > 0 ; idx--)
    {
    *temp_ptr = 0 ;
    temp_ptr++ ;
    }
  
  /* Mark the record image for each maxima found */
  for (idx = SIZE_HASH_TABLE - 1 ; idx >= 0 ; idx--)
    {
    for (tedgels = edgels[idx] ; tedgels != NULL ; tedgels = tedgels->next)
      DEREF(Record_image, XSIZE, tedgels->pos.x, tedgels->pos.y) = MAXIMA ;
    }
  }


/*************************************************************************/
/*                                                                       */
/* JOIN_EDGES                                                            */
/*                                                                       */
/* PURPOSE:                                                              */
/*   This will loop through the structure of high thresholds and find a  */
/*   starting place to look for a valid edge on the image. It will       */
/*   construct a linked list of edges and junctions, where a junction is */
/*   defined to be a place where an edge has more than two neighbors     */
/*   (first in four then in eight connectivity on the compass).  This    */
/*   will return both a list of edges and junctions in the image.        */
/*                                                                       */
/*************************************************************************/
void join_edges(double *gx,	/* X Gradient of convoluted image */
                double *gy)	/* Y Gradient of convoluted image */
  {
  int		bin ;		/* looping index for going through the bins */
  edgel		*edge_s ;	/* Starting pointer of an edge */
  edgel		*edge_e ;	/* Ending pointer of an edge */
  edgel         *reserve ;	/* reserved prev edge pointer */
  junction	*jct ;
  junction	*jct_s ;	/* Starting junction pointer */
  junction	*jct_e ;	/* Ending junction pointer */
  junction_list	*tjct_lst ;	/* Temporary junction list ptr */
  char		neighbor_list[MAXDIRECT] ;	/* A list of possible neighbors */
  int		neighbors ;	/* Number of neighbors in the list*/
  int		idx ;		/* For-loop index */
  int		direct ;	/* Direction of junction */
  int		nx, ny ;	/* New X and Y points used to pick up stray edge points  */
  int		new_edgels_added ;	/* Boolean flag for repeat loop */
  
  /* Initialize the list of edges and junctions to NULL */
  jct_lst = NULL ;		/* This is a global - remember? */
  new_edgels_added = FALSE ;
  
  do
    {
    /* Traverse the linked list of high threshold 'starting' points 
       looking for valid edges to add into the edge list. */
    for (bin = SIZE_HASH_TABLE - 1 ; bin >= 0 ; bin--)
      {
      for (tedgels = edgels[bin] ; tedgels != NULL ; tedgels = tedgels->next)
        {
        /* If this is not a high edge then ignore it */
        if (tedgels->high)
          {
          jct_s = jct_e = NULL ;
          edge_s = edge_e = NULL ;
          
          /* If the point on the image has not been VISITED then perform 
             a recursive walk on the image creating a connected edge that 
             will be inserted into the edge_list */
          if (DEREF(Record_image, XSIZE, 
                    tedgels->pos.x, tedgels->pos.y) == MAXIMA)
            {
            /* preserve the pointer in case of memory screw up */
	    /* modified by Selim
            reserve = tedgels->next ;
	    */
            
            walk(tedgels->pos.x, tedgels->pos.y, NULL, 
                 &edge_s, &edge_e, &jct_s, &jct_e, NONCIRCULAR, TRUE) ;
            
	    /* modified by Selim
            if (tedgels->next != NULL)
              tedgels->next = reserve ;
	    */
            
            /* If an edge has been created, then add it to the edge list */
            if (edge_s != NULL)
              {
              /* Add the edge to the global list called "edges" */
              add_edges(&edges, &edge_s, &edge_e, &jct_s, &jct_e) ;
              
              /* Increment the count of edges */
              NUM_EDGES++;
              }
            }
          } /* end of case for a high edgel */
        } /* end of for-loop through a bin */
      } /* end of loop through all bins */
    
    new_edgels_added = FALSE ;
    
    /* Once we have traversed the list, set any edge points connected
       to junctions that have not yet been visited to "high" and do the
       recursive walk again picking up the values that are above the
       low threshold. */
    for (tjct_lst = jct_lst ; tjct_lst != NULL ; tjct_lst = tjct_lst->next)
      {
      /* Find out how many neighbors of the junction that have not yet 
         been visited. */
      jct = tjct_lst->jct ;
      neighbors = num_neighbors(jct->pos.x, jct->pos.y, 
                                MAXIMA, neighbor_list) ;
      
      /* For all of the neighbors, find the direction that we should look 
         for valid edge points and insert THAT point into the high threshold 
         list. */
      for (idx = 0 ; idx < neighbors ; idx++)
        {
        direct = neighbor_list[idx] ;
        nx = newx(jct->pos.x, idx, neighbor_list) ;
        ny = newy(jct->pos.y, idx, neighbor_list) ;
        
        if (jct->neighbors[direct] == NULL)
          {
          /* Set the associated edgel to high */
          set_edgel_high(edgels, (double) nx, (double) ny) ;
          new_edgels_added = TRUE ;
          }
        else
          DEREF(Record_image, XSIZE, jct->pos.x, jct->pos.y) = 255 ;
        } /* end of loop through the neighbors */
      }
    } while(new_edgels_added == TRUE) ;
  }


/*************************************************************************/
/*                                                                       */
/* ADD_EDGES                                                             */
/*                                                                       */
/* PURPOSE:                                                              */
/*   This will insert into the edge-list-structure all of the pointers   */
/*   necessary to create a linked list of edges. The new edge will be    */
/*   placed on the front of the list.                                    */
/*                                                                       */
/*************************************************************************/
void add_edges(edge **list,		/* An expanded list of valid edge */
               edgel **edge_s,		/* Starting edge pointer */
               edgel **edge_e,		/* Ending edge pointer */
               junction	**jct_s,	/* Starting junction pointer */
               junction	**jct_e)	/* Ending junction pointer */
  {
  edge	*new ;
  
  /* Make room in memory of a new edge */
  new = (edge *) calloc(1, sizeof(edge)) ;
  
  /* Fill the slots in the record */
  new->start_e = *edge_s ;
  new->end_e = *edge_e ;
  new->start_j = *jct_s ;
  new->end_j = *jct_e ;
  new->next = *list ;
  *list = new ;
  }


/*************************************************************************/
/*                                                                       */
/* WALK                                                                  */
/*                                                                       */
/* PURPOSE:                                                              */
/*   This will recursively walk around the image and connect all joining */
/*   edges together until there is either a junction or no more          */
/*   adjoining edge points exist.                                        */
/*                                                                       */
/*************************************************************************/
void walk(double x,		/* X coordinate in the image */
          double y,		/* Y coordinate in the image */
          edgel *prev_edgel,	/* The previous edge point */
          edgel	**start_edgel,	/* Pointer to the start edge point */
          edgel	**end_edgel,	/* Pointer to the end edge point */
          junction **start_jct,	/* Pointers to the beginning and */
          junction **end_jct,	/* the ending junctions */
          int type,		/* CIRCULAR or NONCIRCULAR */
          int add_to_end)
  {
  int		neighbors ;	/* Number of possible neighbors */
  edgel		*e ;		/* New edge point */
  char		neighbor_list[MAXDIRECT] ;	/* Array of neighbor directions */
  junction_list *j ;		/* Junction pointer from jct_lst */
  double	gx, gy;		/* Junction gradient values */
  
  /* Count the number of neighbors that exist for this current point.
     Include all maxima points, as well as the already visited points 
     and junctions.  This way, we can decided if this point is a 
     junction or just should be another visited point. */
  neighbors = num_neighbors(x, y, MAXIMA | VISITED | JUNCTION, neighbor_list) ;
  
  /* Junction */
  if (neighbors > 2)
    {
    neighbors = num_neighbors(x, y, MAXIMA | VISITED, neighbor_list) ;
    
    /* Mark it as a junction */
    DEREF(Record_image, XSIZE, x, y) = JUNCTION ;
    
    /* Find it in the junction list or make a new junction */
    j = get_junction(x, y, gx, gy) ;
    
    if (prev_edgel == NULL)
      *start_jct = *end_jct = j->jct ;
    else
      {
      /* If a previous point exists then set this junction point to
         point to it. */
      set_junction(j, prev_edgel) ;
      *end_jct = j->jct ;
      }
    
    if(junction_done(j, neighbors, neighbor_list))
      delete_junction(j) ;
    } /* End if junction */
  else
    {
    /* Else find out how many unvisited neighbors exist for this point */
    /* Count Maxima and Junctions as unvisited neighbors */
    neighbors = num_neighbors(x, y, MAXIMA | JUNCTION, neighbor_list) ;
    
    /* Mark point as being visited */
    DEREF(Record_image, XSIZE, x, y) = VISITED ;
    
    /* 1st Generation recursion, set all pointers to point to this new 
       edgel we are looking at */
    if(prev_edgel == NULL)
      {
      /* Since this is being called from join_edges, tedgels points to 
         the current edgel.  Make a duplicate edgel. */
      e = (edgel *) calloc(1, sizeof(edgel)) ;
      *e = *tedgels ;
      e->next = e->prev = NULL ;
      *start_edgel = *end_edgel = e ;
      }
    else
      {
      if ( ((*start_edgel)->pos.x == x) && ((*start_edgel)->pos.y == y) )
        {
        type = CIRCULAR ;
        return ;
        }
      else
        {
        e = get_edgel(edgels, x, y) ;
        if(e != NULL)
          {
          if (add_to_end)
            {
            e->prev = prev_edgel ;
            prev_edgel->next = e ;
            e->next = NULL ;
            *end_edgel = e ;
            }
          else
            {
            e->next = prev_edgel ;
            prev_edgel->prev = e ;
            e->prev = NULL ;
            *start_edgel = e ;
            }
          } /* end of case where found the edgel */
        } /* end of case where not circular */
      } /* end else not 1st generation */
    
    if (neighbors == 0)
      return ;
    else
      {
      /* Recursively walk all around and find all my neighbors */
      walk(newx(x, 0, neighbor_list),
           newy(y, 0, neighbor_list),
           e,
           start_edgel, end_edgel,
           start_jct, end_jct, type, add_to_end) ;
      
      /* If there is a neighbor in two different directions, then 
         switch our starting and ending edgels and junctions, so 
         that our code is simplier since we then only have to check 
         the starting points. */
      if (neighbors == 2)
        walk(newx(x, 1, neighbor_list), newy(y, 1, neighbor_list),
             e, start_edgel, end_edgel, end_jct, start_jct, type, FALSE) ;
      }
    }
  }


/*************************************************************************/
/*                                                                       */
/* NEWX                                                                  */
/*                                                                       */
/* PURPOSE:                                                              */
/*   Find the next x direction one should walk based upon the current    */
/*   location and the index-th item in the neighbor list.  Use the       */
/*   x-offsets to make moving in the compass direction easier to figure  */
/*   out.                                                                */
/*                                                                       */
/*************************************************************************/
double newx(double x,			/* Current X location */
            int	index,			/* Index to find the next neighbor */
            char list[MAXDIRECT])	/* List of possible next Neighbors */
  {
  return(x + x_offset[list[index]]) ;
  }


/*************************************************************************/
/*                                                                       */
/* NEWY                                                                  */
/*                                                                       */
/* PURPOSE:                                                              */
/*   Find the next x direction one should walk based upon the current    */
/*   location and the index-th item in the neighbor list.  Use the       */
/*   y-offsets to make moving in the compass direction easier to figure  */
/*   out.                                                                */
/*                                                                       */
/*************************************************************************/
double newy(double y,			/* Current Y location */
            int index,			/* Index to find the next neighbor */
            char list[MAXDIRECT])	/* List of possible next Neighbors */
  {
  return(y + y_offset[list[index]]) ;
  }


/*************************************************************************/
/*                                                                       */
/* NUM_NEIGHBORS                                                         */
/*                                                                       */
/* PURPOSE:                                                              */
/*   Count the number of neighbors that exist in the immediate vacinity. */
/*   That is check North, South, East, and West first. If the two        */
/*   connected directions (i.e. N and E) do not contain a neighbor then  */
/*   check in the combination of those directions, that is, check in the */
/*   North-East corner.  Do this for all four corners.  This way, we     */
/*   check four connectivity before we check eight connectivity.  Type   */
/*   is the type of neighbor we are looking for - it can be either a     */
/*   JUNCTION, or a MAXIMA, or a VISITED point.  Return the count of     */
/*   neighbors and the list of valid directions the neighbors are        */
/*   located at.                                                         */
/*                                                                       */
/*************************************************************************/
int num_neighbors(double x,		/* Current coordinates in the image */
                  double y,
                  int type,
                  char list[MAXDIRECT])	/* A list of possible neighbors and
				           their directions on a compass.  */
  {
  int	n, s, e, w, neighbors ;
  
  neighbors = 0 ;
  
  /* Check to the North */
  if ( (n = DEREF(Record_image, XSIZE, x, y+1)) & type )
    {
    list[neighbors] = N ;
    neighbors++ ;
    }
  
  /* Check to the East */
  if ( (e = DEREF(Record_image, XSIZE, x+1, y)) & type )
    {
    list[neighbors] = E ;
    neighbors++ ;
    }
  
  /* Check to the South */
  if ( (s = DEREF(Record_image, XSIZE, x, y - 1)) & type )
    {
    list[neighbors] = S ;
    neighbors++ ;
    }
  
  /* Check to the West */
  if ( (w = DEREF(Record_image, XSIZE, x-1, y) ) & type )
    {
    list[neighbors] = W ;
    neighbors++ ;
    }
  
  /* Check to the North-East */
  if ( (!(n || e)) && (DEREF(Record_image, XSIZE, x+1, y+1) & type) )
    {
    list[neighbors] = NE ;
    neighbors++ ;
    }
  
  /* Check to the North-West */
  if ( (!(n || w)) && (DEREF(Record_image, XSIZE, x-1, y+1) & type) )
    {
    list[neighbors] = NW ;
    neighbors++ ;
    }
  
  /* Check to the South-East */
  if ( (!(s || e)) && (DEREF(Record_image, XSIZE, x+1, y-1) & type) )
    {
    list[neighbors] = SE ;
    neighbors++ ;
    }
  
  /* Check to the South-West */
  if ( (!(s || w)) && (DEREF(Record_image, XSIZE, x-1, y-1) & type) )
    {
    list[neighbors] = SW ;
    neighbors++ ;
    }
  
  /* Return the number of neighbors we found */
  return neighbors ;
  }


/*************************************************************************/
/*                                                                       */
/* JUNCTION_DONE                                                         */
/*                                                                       */
/* PURPOSE:                                                              */
/*   Since we keep each new junction in a list until we have made sure   */
/*   that all its neighbors have been examined, we need a way to check   */
/*   to see if all the neighbors have been examined.  So this will check */
/*   the eight connectivity of the junction to see what next needs to be */
/*   looked at, and possibly removed from the list.  A junction should   */
/*   be removed from the list if there are no more possible edge points  */
/*   connected to this junction.  This returns a boolean                 */
/*                                                                       */
/*************************************************************************/
int junction_done(junction_list *t_jct_lst,	/* Current junction pointer */
                  int neighbors,		/* Number of neighbors */
                  char n_list[MAXDIRECT])	/* List of neighbors */
  {
  int		idx ;		/* for loop index */
  edgel		*j_edge ;	/* edgel next to a junction */
  
  /* Go around the eight neighbors and if a neighbor is NULL
     or it is NOTUSED then look at the next compass direction */
  for (idx = neighbors - 1 ; idx >= 0 ; idx--)
    {
    j_edge = t_jct_lst->jct->neighbors[n_list[idx]] ;
    if ( (j_edge == NULL) || 
         (DEREF(Record_image, XSIZE, j_edge->pos.x, j_edge->pos.y) & MAXIMA) )
      return FALSE ;
    }
  
  /* If we get here, then all the neighbors have been looked at and 
     this point can be removed from the global jct_lst */
  /* Count the total number of junctions */
  NUM_JCTS++ ;
  
  return TRUE ;
  }


/*************************************************************************/
/*                                                                       */
/* DELETE_JUNCTION                                                       */
/*                                                                       */
/* PURPOSE:                                                              */
/*   Delete the junction current_jct from the global jct_lst             */
/*                                                                       */
/*************************************************************************/
void delete_junction(junction_list *current_jct)	/* Current junction */
  {
  junction_list		*p ;
  junction_list		*temp ;
  
  /* Set temp to the head of the junction list and move it to the next 
     item in the list IF there exists one, until we found a match. P 
     points to the previous pointer. */
  p = NULL ;
  temp = jct_lst ;
  
  while( temp != NULL && temp->jct != NULL && temp != current_jct)
    {
    p = temp ;
    temp = temp->next ;
    }
  
  /* Update the pointers */
  if (p == NULL)
    jct_lst = current_jct->next ;
  else
    p->next = temp->next ;
  }


/*************************************************************************/
/*                                                                       */
/* SET_JUNCTION                                                          */
/*                                                                       */
/* PURPOSE:                                                              */
/*   Mark the junction so that it knows who its neighbors are. Nothing   */
/*   is returned since we are really modifying the global jct_lst.       */
/*                                                                       */
/*************************************************************************/   
void set_junction(junction_list *current_jct,	/* The current junction */
                  edgel *edge_point)	/* The point next to the junction */
  {
  double	x, y ;
  int		idx ;		/* looping index */
  
  /* Find out the direction the edge point is location in regards to 
     the current junction based upon the xoffset and yoffset values. */
  x = edge_point->pos.x - current_jct->jct->pos.x ;
  y = edge_point->pos.y - current_jct->jct->pos.y ;
  
  /* Find the compass direction the edge point is located based upon the 
     x and y offset values */
  for (idx = 0 ; idx < MAXDIRECT; idx++)
    if ( (x == x_offset[idx]) && (y == y_offset[idx]) )
      {
      current_jct->jct->neighbors[idx] = edge_point ;
      return ; 
      }
  }


/*************************************************************************/
/*                                                                       */
/* GET_JUNCTION                                                          */
/*                                                                       */
/* PURPOSE:                                                              */
/*  This will either find a junction in the jct_list based upon its x    */
/*  and y coordinates, or it will create a new junction and put it at    */
/*  the front of the global jct_lst.  This will return a pointer to the  */
/*  junction that should be used next.                                   */
/*                                                                       */
/*************************************************************************/
junction_list *get_junction(double x,	/* X and Y locations of the junction */
                            double y,
                            double gx,	/* The x and y gradient values */
                            double gy)
  {
  junction_list		*temp ;
  junction_list		*t_jct_lst ;
  edgel			*e ;
  int			idx ;		/* looping index */
  
  /* Search the list to see if this junction already exists or not */
  for (temp = jct_lst ; temp != NULL ; temp = temp->next)
    if (temp->jct != NULL && x == temp->jct->pos.x && y == temp->jct->pos.y)
      return temp ;
  
  /* If it doesn't exist, then create it */
  t_jct_lst = (junction_list *) calloc(1, sizeof(junction_list)) ;
  t_jct_lst->jct = (junction *) calloc(1, sizeof(junction)) ;
  
  /* Fill in the record with the known values */
  t_jct_lst->jct->pos.x = x ;
  t_jct_lst->jct->pos.y = y ;
  
  e = get_edgel(edgels, x, y) ;
  
  if (e == NULL)
    {
    t_jct_lst->jct->grad.p = 0.0 ;
    t_jct_lst->jct->grad.q = 0.0 ;
    t_jct_lst->jct->loc.x  = 0.0 ;
    t_jct_lst->jct->loc.y  = 0.0 ;
    }
  else
    {
    t_jct_lst->jct->grad = e->grad ;
    t_jct_lst->jct->loc = e->loc ;
    free(e) ;
    }
  
  /* Set all the junction neighbors to NULL */
  for (idx = 0 ; idx < MAXDIRECT ; idx++)
    t_jct_lst->jct->neighbors[idx] = NULL ;
  
  /* Put it at the front of the jct_list */
  t_jct_lst->next = jct_lst ;
  jct_lst = t_jct_lst ;
  
  return t_jct_lst ;
  } 


/*************************************************************************/
/*                                                                       */
/* GET_EDGEL                                                             */
/*                                                                       */
/* PURPOSE:                                                              */
/*   Find the edgel in the edgel list which has coordinate (x, y).       */
/*   Disconnect the edgel, NULL out its pointer fields, and return a     */
/*   pointer to it.                                                      */
/*                                                                       */
/*************************************************************************/
edgel *get_edgel(edgel *edgel_table[],	/* ptr to the edgel table */
                 double x_pos,	/* X-position of the edgel to retrieve */
                 double y_pos)	/* Y-position of the edgel to retrieve */
  {
  edgel		**bin_ptr ;		/* pointer to the edgel list in the 
                                           appropriate bin */
  edgel		*edgel_ptr ;		/* temp pointer to an edgel */
  edgel		*prev_edgel_ptr ;	/* ptr to a previous edgel */

  /* Start at the list contained in the appropriate bin */
  bin_ptr = &(edgel_table[edgel_key(x_pos, y_pos)]) ;
  
  /* Search this edgel list for a matching edgel */
  for (prev_edgel_ptr = NULL, edgel_ptr = *bin_ptr ;
       edgel_ptr != NULL ;
       edgel_ptr = edgel_ptr->next)
    {
    /* Check for matching x & y coordinates */
    if (edgel_ptr->pos.x == x_pos && edgel_ptr->pos.y == y_pos)
      {
      /* If we have a match, disconnect it, NULL out its pointer fields, 
         and drop out of the loop so that we return a pointer to it */
      if (prev_edgel_ptr == NULL)
        /* The edgel is at the head of the list so re-point the bin pointer */
        *bin_ptr = edgel_ptr->next ;
      else
        /* The desired edgel is not in the front of the bin list */
        prev_edgel_ptr->next = edgel_ptr->next ;
      
      /* NULL out this edgels pointers */
      edgel_ptr->prev = NULL ;
      edgel_ptr->next = NULL ;
      
      /* Break out since found the edgel */
      break ;
      }
    
    /* Current edgel now becomes the previous as the loop continues */
    prev_edgel_ptr = edgel_ptr ;
    }
  
  /* Regardless of what happens, return the pointer. */
  return edgel_ptr ;
  }


/*************************************************************************/
/*                                                                       */
/* GET_EDGEL_PTR                                                         */
/*                                                                       */
/* PURPOSE:                                                              */
/*  Find the edgel in the edgel table which has coordinates (x_pos,      */
/*  y_pos) and return a pointer to it.                                   */
/*                                                                       */
/*************************************************************************/
edgel *get_edgel_ptr(edgel *edgel_table[],	/* ptr to the edgel table */
                     double x_pos,/* X-position of the edgel to retrieve */
                     double y_pos)/* Y-position of the edgel to retrieve */
  {
  edgel		*edgel_ptr ;		/* temp pointer to an edgel */
  
  /* Search this edgel list for the matching edgel */
  for (edgel_ptr = edgel_table[edgel_key(x_pos, y_pos)] ;
       edgel_ptr != NULL ;
       edgel_ptr = edgel_ptr->next)
    /* Check for identical edgel coordinates */
    if (edgel_ptr->pos.x == x_pos && edgel_ptr->pos.y == y_pos)
      break ;		/* Have found a match so break out */

  /* Return the pointer to the edgel */
  return edgel_ptr ;
  }


/*************************************************************************/
/*                                                                       */
/* WRITE FILE                                                            */
/*                                                                       */
/* PURPOSE:                                                              */
/*   Write all of the edges to a file with each edge seperated by two    */
/*   '-1's.  All that the 'next' group needs is the x and y coordinates  */
/*   for each edge point, so that is what we will give them.             */
/*                                                                       */
/*************************************************************************/
void write_file(edge *edge_list,	/* A list of edges */
                char *filename)		/* A file to write the coordinates to */
  {
  FILE		*fp ;
  edge		*tedge_lst ;
  edgel		*e ;
  junction	*j ;
  
  /* Point to the start of the edge list */
  tedge_lst = edge_list ;
  
  if ((fp = fopen(filename, "w")) == NULL)
    {
    fprintf(stderr, "ERROR, can't open file \"%s\" for \"w\"\n", filename) ;
    exit(-1) ;
    }
  
  /* While there are complete edges in the edge_list, loop through
     all the individual edge points  and print it out to a file. */
  while(tedge_lst != NULL)
    {
    /* If there is a junction, write out its x and y position and 
       sub-pixel localized values (if requested) */
    j = tedge_lst->start_j ;
    
    if (j != NULL)
      {
      if (Sub_pixel)
        fprintf(fp, "%.1f %.1f\n", j->loc.x, j->loc.y) ;
      else
        fprintf(fp, "%.2f %.2f\n", j->pos.x, j->pos.y) ;
      }
    
    /* While there is an edge point, write out its x and y position */
    e = tedge_lst->start_e ;
  
    while (e != NULL)
      {
      if (Sub_pixel)
        fprintf(fp, "%.1f %.1f\n", e->loc.x, e->loc.y) ;
      else
        fprintf(fp, "%.2f %.2f\n",e->pos.x,e->pos.y) ;
      e = e->next ;
      }
  
    /* If there is a junction, write out its x and y position and 
       sub-pixel localized values */
    j = tedge_lst->end_j;
  
    if (j != NULL)
      {
      if (Sub_pixel)
        fprintf(fp, "%.1f %.1f\n", j->loc.x, j->loc.y) ;
      else
        fprintf(fp, "%.2f %.2f\n", j->pos.x, j->pos.y) ;
      }
  
    /* Seperate each edge by two -1's */
    fprintf(fp, "-1 -1\n") ;
  
    /* Goto the next edge */
    tedge_lst = tedge_lst->next ;
    }
  
  /* Clean up and quit */
  fclose(fp) ;
  }


/*************************************************************************/
/*                                                                       */
/* LOCALIZE                                                              */
/*                                                                       */
/*  SUB-PIXEL LOCALIZATION FUNCTION                                      */
/*    Consider a point at (x,y) with magnitude G.  Two other magnitudes  */
/*    were computed around (x,y); call these G+ and G-.  Let H denote    */
/*    the distance between (x,y) and the place where G+ is sampled (it   */
/*    is the same for G-).  Ha can range from 1 to (Sqrt 2), depending   */
/*    on where the direction of the gradient at (x,y) was.  If G+ was    */
/*    sampled at (x,y+1), H=1; for G+ at (x+1,y+1), H is (sqrt 2); etc.  */
/*    To determine the exact value for H, it must be computed by         */
/*    Sec(theta) or 1/Cos(theta) where theta is the angle which the      */
/*    direction of G makes with respect to the nearest axis.  A way to   */
/*    detemine H in one step is:                                         */
/*                                                                       */
/*    H = 1/((max (Gx,Gy))/sqrt/(Gx^2+Gy^2))                             */
/*      = sqrt(Gx^2+Gy^2)/max(Gx,Gy)                                     */
/*    Where Gx, Gy are the x and y components of the gradient            */
/*                                                                       */
/*    The goal of sub-pixel localization is to find a better estimate of */
/*    the actual extrema location.  We are going to have to store this   */
/*    these real coordinates somewhere and recall them when we are done  */
/*    processing egdels.  The integer values, which we have now, are     */
/*    necessary for the walk functions to work.                          */
/*                                                                       */
/*    The three values G, G+, G- are assumed to lie on a parabola.  The  */
/*    general form of the parabola is:  g=ah^2 + bh + c.  We want to     */
/*    determine the values of a, b, and c so that we get the exact       */
/*    parabola.  The extremum of this parabola is found by Hm = -b/(2*a) */
/*    by derivative of parabola curve set to 0 and solving for h.        */
/*                                                                       */
/*    Imagine the parabola being the cross section of the gradient       */
/*    magnitudes taken in the direction of G.                            */
/*                                                                       */
/*    G+ = a H^2 + b*H + c  (we know G+ and H already)                   */
/*    G- = a H^2 - b*H + c  (know G-)                                    */
/*    G  = c (since this is at point (x,y), h=0)                         */
/*                                                                       */
/*    These are three linear equations with 3 unknowns                   */
/*                                                                       */
/*    Solving these yields:   b=(G+ - G-)/2H,                            */
/*                            a=(G+ + G- -2G)/2H^2,                      */
/*                            c=G                                        */
/*    (might want to double check the algebra)                           */
/*                                                                       */
/*    Then Hm = -b/(2*a)                                                 */
/*                                                                       */
/*    The real coordinates are then:                                     */
/*                                                                       */
/*    (Xr,Yr) = (x + Hm*(Gx/M) , y + Hm*(Gy/M)) where M=sqrt(Gx^2+Gy^2)  */
/*                                                                       */
/*************************************************************************/
void localize(double g,		/* Center point gradient magnitude squared */
              double gu,	/* Gradient magnitudes squared up and down */
              double gd,
              double direction,	/* Direction of the gradient g */
              double x,		/* X and Y positions of original pixel */
              double y,
              double gx,	/* The gradients in the x and y direction */
              double gy,
              double *xreal,	/* Return the sub pixel coordinate values */
              double *yreal)	/* in terms of x and y */
  {
  double	h ;		/* Distance from g to gu or gd */
  double	displacement ;
  
  if (fabs(direction) < 1.0)
    h = sqrt(1.0 + direction * direction) ;
  else
    h = sqrt(1.0 + 1.0/(direction * direction)) ;
  
  displacement = -(gu - gd) * h / (2 * (gu + gd - 2 * g)) ;
  
  /* If the gradient is pointing down turn it over so that it always
     will point up. */
  if (gx < 0)
    {
    gx = -gx ;
    gy = -gy ;
    }
  
  *xreal = x + displacement * gx / g ;
  *yreal = y + displacement * gy / g ;
  }


/*************************************************************************/
/*                                                                       */
/* NEW_EDGEL_LIST                                                        */
/*                                                                       */
/*************************************************************************/
edgel **new_edgel_list()
  {
  edgel		**edgel_table ;		/* hash table for edgels */
  int		idx ;			/* looping index */

  /* Allocate space for the edgel table */
  edgel_table = (edgel **) calloc(SIZE_HASH_TABLE, sizeof(edgel *)) ;

  /* Initialize all the entries to NULL */
  for (idx = SIZE_HASH_TABLE - 1 ; idx >= 0 ; idx--)
    edgel_table[ idx ] = NULL ;

  /* Finished OK so return the ptr to the hash table */
  return( edgel_table ) ;
  } 


/*************************************************************************/
/*                                                                       */
/* EDGEL_KEY                                                             */
/*                                                                       */
/* Purpose:                                                              */
/*   Calculate a key (index into the hash table) using the given         */
/*   position of the edgel.                                              */
/*                                                                       */
/*************************************************************************/
int edgel_key(double x_pos,	/* X-position of the edgel */
              double y_pos)	/* Y-position of the edgel */
  {
  int	key ;			/* calculated key value */
  
  /* Value of the key is the of the siz digit number cat( X, Y )
     MOD the size of the hash-table */
  key = ((int) (x_pos * 1000 + y_pos)) % SIZE_HASH_TABLE ;
  
  /* Finished OK so return the key value */
  return key ;
  }
